
/********************************************************************
 *
 * FILE:
 *	stNewAPIChorus.c
 *
 * DESCRIPTION:
 *	Uses the API to link up to the mix engine. The API does the necessary
 *	connections to shared memory, and does a simple negotiation with the
 *	frontend to request addition to the EffectSLab frontend. The APIs are
 *	really pretty trivial, in as much as there is no flexibility the what
 *	you can do, just attach, initialise, and do read/write operations. There
 *	are no bells or whistles. There aren't even any manual pages. The up
 *	side of this approach is simplicity - the user is removed from the details
 *	of SHMEM and the specifics of the mixSLab internal memory formats. The
 *	obvious downside is the lack of flexibility.
 *	Since there are is little *	documentation on the APIs, here the details:
 *
 *		int * seFullAttach(int memID, int busID, int channels,
 *			int (*dspAlgo)(), int (*paramAlgo)(),
 *			char *name, char *shortName,
 *			char *control0, char *control1,
 *			char *control2, char *control3,
 *			char *control4, char *control5,
 *			char *control6, char *control7)
 *
 *			memId is the shared memory segment for the controlBuffer, and
 *			busID is the bus to which to attach. These are passed as argv[1]
 *			and argv[2] when the process is forked by the frontend. Channels
 *			is SLAB_EFFECTS_STEREO or SLAB_EFFECTS_MONO. dspAlgo() is called
 *			when data is available, paramAlgo() is called when a parameter is
 *			changed. Name and shortName are used in the frontend display, and
 *			the	null terminated list of parameters are for the display, max 8.
 *
 *		int seGetCtrlVal(int controllerID)
 *
 *			The controllerID is implied by the call to seFullAttach().
 *			Returns the current value of the given controller, as set by
 *			the user in the EffectSLab gui.
 *
 *		int seSetCtrlDisp(int controllerID, char *dispText)
 *
 *			The controllerID implied by a previous seFullAttach() call.
 *			Sets a string to be displayed in the
 *			frontend for the value of this controller. This will allow the
 *			effect to set actual values and units in the display.
 *
 * TOC:
 *	main()
 *	chorus()
 *	checkParams()
 *
 *******************************************************************/

#include "slabEffects.h"

extern int sampleRate, resolution;

/*
 * Used by the user interface in the effects rack.
 */
#define EFFECT_NAME "stNewAPIChorus"
#define SHORT_NAME "Chorus"

/*
 * Preassigned indeces for the controllers. The values reflect the location of
 * the parameter definition in the call to seFullAttach().
 */
#define SPEED		0
#define BREADTH		1

/*
 * LOCAL VARIABLE DEFINITIONS - data required by your DSP algorithm.
 *
 * These are dependant on the effect - parameters etc. This chorus implements
 * 4 individual delay lines, with a MAX delay, at 16384 samples, or about
 * 400ms at 44100sps.
 */
#define HISTORY_SIZE 65536
#define CHORUS_SIZE 16384
#define INDEX1_START 0
#define INDEX2_START 16384
#define INDEX3_START 32768
#define INDEX4_START 49152

int history[HISTORY_SIZE];
int indexOut1 = 0;
int indexOut2 = 0;
int indexOut3 = 0;
int indexOut4 = 0;
int indexIn = 0;
int Speed1, Speed2, Speed3, Speed4;

int speed, breadth;
int Breadth = 128;

int currentBreadth1 = 0;
int currentBreadth2 = 0;
int currentBreadth3 = 0;
int currentBreadth4 = 0;
int countBreadth1 = 0;
int countBreadth2 = 0;
int countBreadth3 = 0;
int countBreadth4 = 0;
int changeBreadth1 = 1;
int changeBreadth2 = 1;
int changeBreadth3 = 1;
int changeBreadth4 = 1;

int Min = 1;

int checkParams(), chorus();

/*
 * When the user interface attaches an effect, it passes two parameters, no
 * switches. This is very rudimentary, but will be extended:
 *
 * argv[1] is the controlbuffer segID, argv[2] should be the bus requested
 */
main(argc, argv)
int argc;
char *argv[];
{
	if (argc < 3)
	{
		printf("%s: parameter problem\n", argv[0]);
		exit(-1);
	}

	/*
	 * Any local initialisation here.
	 */

	/*
	 * seFullAttach() will register with the mixengine, and link up the shared
	 * memory control and data structures. It effectively reserves a bus.
	 * The SLAB_EFFECTS_MONO really refers to the first bus attachment. We
	 * can attach to a second one with the SLAB_EFFECTS_STEREO flag. It is
	 * only used as a switch to use a slightly different link up protocol.
	 *
	 * The checkParams() is called when one of the params changes.
	 * dspAlgo chorus() is called when data is available.
	 */
	if ((seFullAttach(atoi(argv[1]), atoi(argv[2]), SLAB_EFFECTS_STEREO,
		chorus, checkParams, /* Code sections */
		EFFECT_NAME, SHORT_NAME, /* GUI identifiers */
		"Speed", "Breadth", (char *) NULL)) /* Controllers */
		!= 0)
		printf("%s: could not attach to requested bus\n", argv[0]);

	/*
	 * This last call, if correct, should never return except if we are 
	 * forced to detach from the GUI.
	 *
	 * Any local cleanup here.
	 */

	exit(0);
}

/*
 * This is called by the library when a parameter is changed. Since the same
 * callback method is used by both library access mechanisms, we are passed
 * two parameters. The first is superfluous for this access mechanism, the 
 * second is an * indication of which parameter changed.
 */
checkParams(devPtr, param)
int *devPtr, param;
{
	char text[32];

	if (param == SPEED)
	{
		Speed1 = (256 - seGetCtrlValByID(SPEED)) << 2;
		Speed2 = Speed1 + 10;
		Speed3 = Speed1 + 20;
		Speed4 = Speed1 + 30;

		/*
		 * The actual speed here is a function of both Speed and Breadth.
		 */
		sprintf(text, "Speed: %2.2f Hz",
			(float) sampleRate / (Speed1 * Breadth));
		seSetCtrlDispByID(SPEED, text);
	}

	Breadth = (seGetCtrlValByID(BREADTH) << 4) + 1;

	sprintf(text, "Breadth: %2.2f s", (float) Breadth/sampleRate);
	seSetCtrlDispByID(BREADTH, text);
}

/*
 * This is the actual chorus algorithm. The rest of this file, excluding the
 * variable definitions, is basically an effects template. If we want to have
 * a different stereo effect, we need to change the algorithm below.
 *
 * If you want to code some of your own effects, you do not need to understanc
 * this algorithm. What you need to understand is the call to seFullAttach(),
 * and the checkParams() subroutine. Note that this effect only uses the 
 * second bus as a return channel, no data is taken off it, only put back.
 * Secondly, even if this is a MONO effect, the DSP algorithm is called with
 * busData and busData2, but the latter will be either NULL or -1.
 */
chorus(busData, busData2, count)
register int *busData, *busData2, count;
{
	register int holder;

	for (; count > 0; count--)
	{
		history[indexIn] = 
		history[indexIn + INDEX2_START] =
		history[indexIn + INDEX3_START] =
		history[indexIn + INDEX4_START] = *busData;

		/*
		 * Fill the second output bus - overwrite contents
		 */
		*(busData2++) = history[indexOut1] +
			history[indexOut3 + INDEX3_START];

		/*
		 * Similarly the first bus
		 */
		*(busData++) = history[indexOut2 + INDEX2_START] +
			history[indexOut4 + INDEX4_START];

		if (++countBreadth1 >= Speed1)
		{
			countBreadth1 = 0;

			if ((currentBreadth1 += changeBreadth1) >= Breadth)
				changeBreadth1 = -1;
			if (currentBreadth1 <= Min)
				changeBreadth1 = 1;

			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth1 < 0) {
				if ((indexOut1 += 1) >= Breadth) indexOut1 = 0;
				if ((indexOut1 += 1) >= Breadth) indexOut1 = 0;
			}
		} else {
			if (++indexOut1 >= Breadth) indexOut1 = 0;
		}

		if (++countBreadth2 >= Speed2)
		{
			countBreadth2 = 0;

			if ((currentBreadth2 += changeBreadth2) >= Breadth)
				changeBreadth2 = -1;
			if (currentBreadth2 <= Min)
				changeBreadth2 = 1;
			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth2 < 0) {
				if ((indexOut2 += 1) >= Breadth) indexOut2 = 0;
				if ((indexOut2 += 1) >= Breadth) indexOut2 = 0;
			}
		} else {
			if (++indexOut2 >= Breadth) indexOut2 = 0;
		}

		if (++countBreadth3 >= Speed3)
		{
			countBreadth3 = 0;

			if ((currentBreadth3 += changeBreadth3) >= Breadth)
				changeBreadth3 = -1;
			if (currentBreadth3 <= Min)
				changeBreadth3 = 1;

			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth3 < 0) {
				if ((indexOut3 += 1) >= Breadth) indexOut3 = 0;
				if ((indexOut3 += 1) >= Breadth) indexOut3 = 0;
			}
		} else {
			if (++indexOut3 >= Breadth) indexOut3 = 0;
		}

		if (++countBreadth4 >= Speed4)
		{
			countBreadth4 = 0;

			if ((currentBreadth4 += changeBreadth4) >= Breadth)
				changeBreadth4 = -1;
			if (currentBreadth4 <= Min)
				changeBreadth4 = 1;

			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth4 < 0) {
				if ((indexOut4 += 1) >= Breadth) indexOut4 = 0;
				if ((indexOut4 += 1) >= Breadth) indexOut4 = 0;
			}
		} else {
			if (++indexOut4 >= Breadth) indexOut4 = 0;
		}

		if (++indexIn >= Breadth) indexIn = 0;
	} 
}

